from pathlib import Path

from django.core.management.base import BaseCommand, CommandError

from ...components import convert_to_camel_case, convert_to_snake_case


COMPONENT_FILE = """
from django_unicorn.components import Component


class {camel_case_component_name}(Component):
    pass

"""

TEMPLATE_FILE = """
<div>
    <!-- put code here -->
</div>

"""


class Command(BaseCommand):
    help = "Creates a new component for `django-unicorn`"

    def add_arguments(self, parser):
        parser.add_argument("component_names", nargs="+", type=str)

    def handle(self, *args, **options):
        if not Path("manage.py").exists():
            raise CommandError("Can't find manage.py in current path.")

        if not Path("unicorn").exists():
            Path("unicorn").mkdir()
            self.stdout.write(self.style.SUCCESS("Create unicorn directory ✨"))

        for component_name in options["component_names"]:
            snake_case_component_name = convert_to_snake_case(component_name)
            camel_case_component_name = convert_to_camel_case(component_name)

            # Create component
            if not Path("unicorn/components").exists():
                Path("unicorn/components").mkdir()

            component_path = Path(f"unicorn/components/{snake_case_component_name}.py")
            component_path.write_text(
                COMPONENT_FILE.format(
                    **{"camel_case_component_name": camel_case_component_name}
                )
            )
            self.stdout.write(self.style.SUCCESS(f"Created {component_path}."))

            # Create template
            if not Path("unicorn/templates/unicorn").exists():
                Path("unicorn/templates/unicorn").mkdir()

            template_path = Path(f"unicorn/templates/unicorn/{component_name}.html")
            template_path.write_text(TEMPLATE_FILE)
            self.stdout.write(self.style.SUCCESS(f"Created {template_path}."))



from django import template
from django.conf import settings

from ..components import Component


register = template.Library()


@register.inclusion_tag("unicorn/scripts.html")
def unicorn_scripts():
    return {"debug": settings.DEBUG}


@register.inclusion_tag("unicorn/styles.html")
def unicorn_styles():
    return {}


def unicorn(parser, token):
    try:
        (tag_name, component_name) = token.split_contents()
    except ValueError:
        raise template.TemplateSyntaxError(
            "%r tag requires a single argument" % token.contents.split()[0]
        )

    if not (
        component_name[0] == component_name[-1] and component_name[0] in ('"', "'")
    ):
        raise template.TemplateSyntaxError(
            "%r tag's argument should be in quotes" % tag_name
        )

    return UnicornNode(component_name[1:-1])


class UnicornNode(template.Node):
    def __init__(self, component_name):
        self.component_name = component_name

    def render(self, context):
        component = Component.create(self.component_name)
        rendered_component = component.render()

        return rendered_component


register.tag("unicorn", unicorn)



import hmac
import importlib
import inspect
import uuid

from django.conf import settings
from django.template import Context
from django.template.engine import Engine
from django.utils.safestring import mark_safe

import orjson
from bs4 import BeautifulSoup
from bs4.formatter import HTMLFormatter


def convert_to_snake_case(s):
    # TODO: Better handling of dash->snake
    return s.replace("-", "_")


def convert_to_camel_case(s):
    # TODO: Better handling of dash/snake->camel-case
    s = convert_to_snake_case(s)
    return "".join(word.title() for word in s.split("_"))


class Component:
    def __init__(self, component_name, id=None):
        if not id:
            id = uuid.uuid4()

        self.id = id
        self.component_name = component_name

    @staticmethod
    def create(component_name, id=None):
        # TODO: Handle the module not being found
        module_name = convert_to_snake_case(component_name)
        module = importlib.import_module(f"unicorn.components.{module_name}")

        # TODO: Handle the class not being found
        class_name = convert_to_camel_case(module_name)
        component_class = getattr(module, class_name)

        return component_class(component_name, id=id)

    def __attributes__(self):
        """
        Get attributes that can be called in the component.
        """
        non_callables = [
            member[0] for member in inspect.getmembers(self, lambda x: not callable(x))
        ]
        attribute_names = list(
            filter(lambda name: self._is_public_name(name), non_callables,)
        )

        attributes = {}

        for attribute_name in attribute_names:
            attributes[attribute_name] = object.__getattribute__(self, attribute_name)

        return attributes

    def __methods__(self):
        """
        Get methods that can be called in the component.
        """

        # TODO: Should only take methods that only have self argument?
        member_methods = inspect.getmembers(self, inspect.ismethod)
        public_methods = filter(
            lambda method: self._is_public_name(method[0]), member_methods
        )
        methods = {k: v for (k, v) in public_methods}

        return methods

    def __context__(self):
        """
        Collects every thing that could be used in the template context.
        """
        return {
            "attributes": self.__attributes__(),
            "methods": self.__methods__(),
        }

    def render(self, template_context={}, include_component_init=True):
        context = self.__context__()
        context_variables = {}
        context_variables.update(context["attributes"])
        context_variables.update(context["methods"])
        context_variables.update(template_context)

        frontend_context_variables = {}
        frontend_context_variables.update(context["attributes"])
        frontend_context_variables = orjson.dumps(frontend_context_variables).decode(
            "utf-8"
        )

        checksum = hmac.new(
            str.encode(settings.SECRET_KEY),
            str.encode(frontend_context_variables),
            digestmod="sha256",
        ).hexdigest()

        if settings.DEBUG:
            context_variables.update({"unicorn_debug": context})

        template_engine = Engine.get_default()
        # TODO: Handle looking in other directories for templates
        template = template_engine.get_template(f"unicorn/{self.component_name}.html")
        context = Context(context_variables, autoescape=True)
        rendered_template = template.render(context)

        soup = BeautifulSoup(rendered_template, features="html.parser")
        root_element = Component._get_root_element(soup)
        root_element["unicorn:id"] = str(self.id)
        root_element["unicorn:checksum"] = checksum

        if include_component_init:
            script = soup.new_tag("script")
            init = {
                "id": str(self.id),
                "name": self.component_name,
            }
            init = orjson.dumps(init).decode("utf-8")
            script.string = f"Unicorn.setData({frontend_context_variables}); Unicorn.componentInit({init});"
            root_element.insert_after(script)

        rendered_template = Component._desoupify(soup)
        rendered_template = mark_safe(rendered_template)

        return rendered_template

    def _is_public_name(self, name):
        """
        Determines if the name should be sent in the context.
        """
        protected_names = (
            "id",
            "render",
            "view",
        )
        excludes = []

        if hasattr(self, "Meta") and hasattr(self.Meta, "exclude"):
            excludes = self.Meta.exclude

        return not (name.startswith("_") or name in protected_names or name in excludes)

    @staticmethod
    def _get_root_element(soup):
        for element in soup.contents:
            if element.name and element.name == "div":
                return element

        raise Exception("No root element found")

    @staticmethod
    def _desoupify(soup):
        soup.smooth()
        return soup.encode(formatter=UnsortedAttributes()).decode("utf-8")


class UnsortedAttributes(HTMLFormatter):
    """
    Prevent beautifulsoup from re-ordering attributes.
    """

    def attributes(self, tag):
        for k, v in tag.attrs.items():
            yield k, v



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_unicorn',
 'django_unicorn.management.commands',
 'django_unicorn.templatetags']

package_data = \
{'': ['*'],
 'django_unicorn': ['static/img/*',
                    'static/js/.babelrc',
                    'static/js/.babelrc',
                    'static/js/.babelrc',
                    'static/js/.babelrc',
                    'static/js/.babelrc',
                    'static/js/0.1.0/*',
                    'static/js/debounce.js',
                    'static/js/debounce.js',
                    'static/js/debounce.js',
                    'static/js/debounce.js',
                    'static/js/debounce.js',
                    'static/js/morphdom/2.6.1/morphdom-livewire.js',
                    'static/js/morphdom/2.6.1/morphdom-livewire.js',
                    'static/js/morphdom/2.6.1/morphdom-livewire.js',
                    'static/js/morphdom/2.6.1/morphdom-umd.js',
                    'static/js/morphdom/2.6.1/morphdom-umd.js',
                    'static/js/morphdom/2.6.1/morphdom-umd.js',
                    'static/js/morphdom/2.6.1/morphdom-umd.min.js',
                    'static/js/morphdom/2.6.1/morphdom-umd.min.js',
                    'static/js/morphdom/2.6.1/morphdom-umd.min.js',
                    'static/js/unicorn.js',
                    'static/js/unicorn.js',
                    'static/js/unicorn.js',
                    'static/js/unicorn.js',
                    'static/js/unicorn.js',
                    'static/js/unicorn.min.js',
                    'static/js/unicorn.min.js',
                    'static/js/unicorn.min.js',
                    'static/js/unicorn.min.js',
                    'static/js/unicorn.min.js',
                    'static/js/utils.js',
                    'static/js/utils.js',
                    'static/js/utils.js',
                    'static/js/utils.js',
                    'static/js/utils.js',
                    'templates/unicorn/*']}

install_requires = \
['beautifulsoup4>=4.9.1,<5.0.0', 'django>=3.0.0', 'orjson>=3.2.1,<4.0.0']

setup_kwargs = {
    'name': 'django-unicorn',
    'version': '0.1.0',
    'description': 'Magical fullstack framework for Django.',
    'long_description': None,
    'author': 'Adam Hill',
    'author_email': 'unicorn@adamghill.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import hmac

from django.conf import settings
from django.http import JsonResponse

import orjson

from .components import Component


def message(request, component_name):
    body = orjson.loads(request.body)
    data = body.get("data", {})
    checksum = body.get("checksum")

    generated_checksum = hmac.new(
        str.encode(settings.SECRET_KEY), orjson.dumps(data), digestmod="sha256",
    ).hexdigest()

    if checksum != generated_checksum:
        return JsonResponse({"error": "Checksum does not match"})

    unicorn_id = body.get("id")
    component = Component.create(component_name, id=unicorn_id)

    for (name, value) in data.items():
        if hasattr(component, name):
            setattr(component, name, value)

    action_queue = body.get("actionQueue", [])

    for action in action_queue:
        action_type = action.get("type")
        payload = action.get("payload", {})

        if action_type == "syncInput":
            name = payload.get("name")
            value = payload.get("value")

            if name is not None and value is not None and hasattr(component, name):
                setattr(component, name, value)

            data[name] = value
        elif action_type == "callMethod":
            name = payload.get("name")
            params = []

            if "(" in name and ")" in name:
                param_idx = name.index("(")
                params = name[param_idx:]

                # Remove the arguments from the method name used later
                name = name.replace(params, "")

                # Remove paranthesis
                params = params[1:-1]

                # Rmeove extra quotes for strings
                if params.startswith("'") and params.endswith("'"):
                    params = params[1:-1]
                elif params.startswith('"') and params.endswith('"'):
                    params = params[1:-1]

                # TODO: Handle kwargs
                params = params.split(",")

            if name is not None and hasattr(component, name):
                func = getattr(component, name)

                if params:
                    func(*params)
                else:
                    func()

                for (
                    attribute_name,
                    attribute_value,
                ) in component.__attributes__().items():
                    data[attribute_name] = attribute_value

    rendered_component = component.render(include_component_init=False)

    res = {
        "id": unicorn_id,
        "dom": rendered_component,
        "data": data,
    }

    return JsonResponse(res)



from django.urls import path

from . import views

app_name = "django_unicorn"


urlpatterns = (
    path("message/<str:component_name>", views.message, name="message"),
    path(
        "message", views.message, name="message"
    ),  # Only here to build the correct url in scripts.html
)
